{*************************************************************************************************}
{*************************************************************************************************}
{**                                                                                             **}
{**                                IC Kommunikation ber USI                                   **}
{**                                                                                             **}
{*************************************************************************************************}
{*************************************************************************************************}

Unit USITWI;

interface

{*************************************************************************************************}
{ Var Declarations                                                                                }
{*************************************************************************************************}
{$IDATA}
  var
    RxCount: byte;
    TxCount: byte;                                       //Pufferfllstnde

{*************************************************************************************************}
{ Functions/Procedures                                                                            }
{*************************************************************************************************}
  procedure TWIInitSlave(Address: Byte);                 //TWI-Slave initialisieren
  function TWIStat: boolean;                             //Abfrage ob Empfangsbyte verfgbar sind
  function TWIIn: byte;                                  //Empfangsbyte lesen
  procedure TWIOut(Data: byte);                          //Daten senden

implementation

{*************************************************************************************************}
{ Type Declarations                                                                               }
{*************************************************************************************************}
  type
{$ENUMTOASM}
    TTWISTate = (usCHECKADDR,usSENDDATA,usREQREPLYSEND,usCHECKREPLYSEND,usREQDATA,usGETDATAACK);
  
{*************************************************************************************************}
{ Const Declarations                                                                              }
{*************************************************************************************************}
  const
{$IFDEF TINY861}
    USIPort: byte = $38;                                 //Port B
    USIOut: byte = USIPort;                              //Portregister
    USIInp: byte = USIPort-2;                            //Pinregister
    USIDDR: byte = USIPort-1;                            //Datenrichtungsregister
    SDAPin: byte = 0;                                    //Datenpin
    SCLPin: byte = 2;                                    //Taktpin
{$ENDIF}
// -- Bit in USISR --
    _USISIF: byte = 7;                                   //Start Condition Interrupt Flag
    _USIOIF: byte = 6;                                   //Counter Overflow Interrupt Flag
    _USIPF: byte = 5;                                    //Stop Condition Flag
    _USIDC: byte = 4;                                    //Data Output Collision
    _USIICNT3: byte = 3;                                 //Counter Value, Bit 3
    _USIICNT2: byte = 2;                                 //Counter Value, Bit 2
    _USIICNT1: byte = 1;                                 //Counter Value, Bit 1
    _USIICNT0: byte = 0;                                 //Counter Value, Bit 0
    USISIF: byte = 1 shl _USISIF;
    USIOIF: byte = 1 shl _USIOIF;
    USIPF: byte = 1 shl _USIPF;
    USIDC: byte = 1 shl _USIDC;
    USIICNT3: byte = 1 shl _USIICNT3;
    USIICNT2: byte = 1 shl _USIICNT2;
    USIICNT1: byte = 1 shl _USIICNT1;
    USIICNT0: byte = 1 shl _USIICNT0;
// -- Bit in USICR --
    _USISIE: byte = 7;                                   //Start Condition Interrupt Enable
    _USIOIE: byte = 6;                                   //Counter Overflow Interrupt Enable
    _USIWM1: byte = 5;                                   //Wire Mode
    _USIWM0: byte = 4;
    _USICS1: byte = 3;                                   //Clock Source Select
    _USICS0: byte = 2;
    _USICLK: byte = 1;                                   //Clock Strobe
    _USITC: byte = 0;                                    //Toggle Clock Port Pin
    USISIE: byte = 1 shl _USISIE;
    USIOIE: byte = 1 shl _USIOIE;
    USIWM1: byte = 1 shl _USIWM1;
    USIWM0: byte = 1 shl _USIWM0;
    USICS1: byte = 1 shl _USICS1;
    USICS0: byte = 1 shl _USICS0;
    USICLK: byte = 1 shl _USICLK;
    USITC: byte = 1 shl _USITC;

{*************************************************************************************************}
{ Var Declarations                                                                                }
{*************************************************************************************************}
  var
{$IDATA}
    iTWIAddress: byte;                                   //eigene Adresse
    oTWIState: TTWIState;                                //Status
    oRxBuffer: array[0..RxBuffLen-1] of byte;            //Empfangspuffer
    oTxBuffer: array[0..TxBuffLen-1] of byte;            //Sendepuffer
    iRxWrPtr: byte;                                      //Schreibzeiger Empfangspuffer
    iRxRdPtr: byte;                                      //Lesezeiger Empfangspuffer
    iTxWrPtr: byte;                                      //Schreibzeiger Sendepuffer
    iTxRdPtr: byte;                                      //Lesezeiger Sendepuffer

{*************************************************************************************************}
{ Functions/Procedures                                                                            }
{*************************************************************************************************}
{$IDATA}
{-------------------------------------------------------------------------------------------------}
{  Puffer leeren                                                                                  }
{-------------------------------------------------------------------------------------------------}
  procedure USIFlushTWIBuffers;
  begin
    asm;
      clr   _ACCA
      ldi   _ACCCLO,USITWI.iRxWrPtr and 0FFH
      ldi   _ACCCHI,USITWI.iRxWrPtr shrb 8
      st    Z+,_ACCA                                     ;Pufferzeiger auf 0
      st    Z+,_ACCA
      st    Z+,_ACCA
      st    Z+,_ACCA
      ldi   _ACCCLO,USITWI.RxCount and 0FFH
      ldi   _ACCCHI,USITWI.RxCount shrb 8
      st    Z+,_ACCA                                     ;Puffer leer
      st    Z+,_ACCA
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  TWI-Slave initialisieren                                                                       }
{-------------------------------------------------------------------------------------------------}
  procedure TWIInitSlave(Address: Byte);
  begin
    asm;
      call  USITWI.USIFlushTWIBuffers                    ;Puffer leeren
      cbi   USIPP,0                                      ;USI auf PB0..PB2
      ld    _ACCA,Y
      andi  _ACCA,07FH                                   ;Adressbereich 1..127
      sts   USITWI.iTWIAddress,_ACCA                     ;Adresse eintragen
      sbi   USITWI.USIOut,USITWI.SCLPin                  ;SCL auf H
      sbi   USITWI.USIOut,USITWI.SDAPin                  ;SDA auf H
      sbi   USITWI.USIDDR,USITWI.SCLPin                  ;SCL als Ausgang
      cbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Eingang
      ldi   _ACCA,USITWI.USISIE+USITWI.USIWM1+USITWI.USICS1
      out   USICR,_ACCA                                  ;TWI-Mode, externer Takt, Start-Interrupt
      ldi   _ACCA,USITWI.USISIF+USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC
      out   USISR,_ACCA                                  ;Flags lschen
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Abfragen, ob Zeichen empfangen wurde                                                           }
{-------------------------------------------------------------------------------------------------}
{$NORETURNCHECK}
  function TWIStat: boolean;
  begin
    asm;
      clr   _ACCA
      lds   _ACCB,USITWI.RxCount                         ;Zeichen im Lesepuffer ?
      tst   _ACCB
      breq  twistat_end                                  ;nein
      ser   _ACCA
twistat_end:
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Zeichen aus Empfangspuffer lesen                                                               }
{-------------------------------------------------------------------------------------------------}
{$NORETURNCHECK}
  function TWIIn: byte;
  begin
    asm;
twiinp_loop:
      lds   _ACCA,USITWI.RxCount                         ;Zeichen im Lesepuffer ?
      tst   _ACCA
      breq  twiinp_loop                                  ;nein, warten
      cli                                                ;Interrupts sperren
      lds   _ACCB,USITWI.iRxRdPtr                        ;Lesezeiger
      clr   _ACCA
      ldi   _ACCCLO,USITWI.oRxBuffer AND 0FFH
      ldi   _ACCCHI,USITWI.oRxBuffer SHRB 8              ;Pufferanfang
      add   _ACCCLO,_ACCB
      adc   _ACCCHI,_ACCA                                ;Leseposition
      ld    _ACCA,Z                                      ;aus Puffer
      inc   _ACCB                                        ;Zeiger weiterrcken
      cpi   _ACCB,RxBuffLen                              ;Pufferende ?
      brne  twiinp_ok                                    ;nein
      clr   _ACCB                                        ;wieder auf Anfang
twiinp_ok:
      sts   USITWI.iRxRdPtr,_ACCB                        ;Pufferzeiger merken
      lds   _ACCB,USITWI.RxCount                         ;Zeichenzhler
      dec   _ACCB                                        ;erniedrigen
      cpi   _ACCB,0FFH                                   ;Unterlauf
      breq  twiinp_end                                   ;ja, verhindern
      sts   USITWI.RxCount,_ACCB                         ;merken
twiinp_end:
      sbrc  Flags,IntFlag
      sei                                                ;Interrupts freigeben
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Zeichen in Sendepuffer schreiben                                                               }
{-------------------------------------------------------------------------------------------------}
{$DEVICE}
  procedure TWIOut(Data: byte);
  begin
    asm;
twiout_loop:
      lds   _ACCB,USITWI.TxCount                         ;Fllstand Sendepuffer
      cpi   _ACCB,TxBuffLen                              ;voll ?
      breq  twiout_loop                                  ;ja, warten
      cli                                                ;Interrupts sperren
      lds   _ACCB,USITWI.iTxWrPtr                        ;Schreibzeiger
      ldi   _ACCCLO,USITWI.oTxBuffer AND 0FFH
      ldi   _ACCCHI,USITWI.oTxBuffer SHRB 8              ;Pufferanfang
      add   _ACCCLO,_ACCB
      clr   _ACCB
      adc   _ACCCHI,_ACCB                                ;Schreibposition
      st    Z,_ACCA                                      ;eintragen
      lds   _ACCB,USITWI.TxCount                         ;Fllstand Sendepuffer
      inc   _ACCB                                        ;+1
      sts   USITWI.TxCount,_ACCB
      lds   _ACCB,USITWI.iTxWrPtr                        ;Schreibzeiger
      inc   _ACCB                                        ;weiterrcken
      cpi   _ACCB,TxBuffLen                              ;Pufferende ?
      brne  twiout_ok                                    ;nein
      clr   _ACCB                                        ;wieder auf Anfang
twiout_ok:
      sts   USITWI.iTxWrPtr,_ACCB                        ;Pufferzeiger merken
      sbrc  Flags,IntFlag
      sei                                                ;Interrupts freigeben
    endasm;
  end;
{*************************************************************************************************}
{ Interrupt-Routines                                                                              }
{*************************************************************************************************}
{$IDATA}
{-------------------------------------------------------------------------------------------------}
{ USI Startbedingung TWI erkannt                                                                  }
{-------------------------------------------------------------------------------------------------}
{$NOSAVE}
  interrupt USISTART;
  begin
    asm;
      sbi   USISR,USITWI._USIPF                          ;Stopbedingungsflag lschen
      ldi   _ACCA,USITWI.usCHECKADDR
      sts   USITWI.oTWIState,_ACCA                       ;warte auf Adresse
      cbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Eingang
usistart_loop:
      sbic  USISR,USITWI._USIPF                          ;Stopbedingung erkannt
      rjmp  usistart_end                                 ;beenden
      sbic  USITWI.USIInp,USITWI.SCLPin                  ;SCL low ?
      rjmp  usistart_loop                                ;nein, warten
usistart_end:
      ldi   _ACCA,USITWI.USISIE+USITWI.USIOIE+USITWI.USIWM1+USITWI.USICS1
      out   USICR,_ACCA                                  ;TWI-Mode, externer Takt, Start-Interrupt, berlauf-Interrupt
      ldi   _ACCA,USITWI.USISIF+USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC
      out   USISR,_ACCA                                  ;Flags lschen, Zhler auf 0
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{ Zhler im USI bergelaufen                                                                      }
{-------------------------------------------------------------------------------------------------}
{$NOSAVE}
  interrupt USIOVF;
  begin
    asm;
      lds   _ACCA,USITWI.oTWIState;                      ;aktueller Zustand
      cpi   _ACCA,USITWI.usCHECKADDR                     ;warte auf Adresse ?
      brne  usiovf_next1                                 ;nein
      in    _ACCA,USIDR                                  ;Adresse holen
      tst   _ACCA                                        ;0 ?
      breq  usiovf_addr                                  ;ja Broadcast
      lds   _ACCB,USITWI.iTWIAddress                     ;eigene Adresse
      lsl   _ACCB                                        ;schieben wegen Richtungsbyte
      eor   _ACCB,_ACCA                                  ;vergleichen
      andi  _ACCB,0FEH                                   ;Richtungsbyte ist egal
      breq  usiovf_addr                                  ;ja, Adresse ist richtig
usiovf_startcond:                                        ;nein, auf neue Startbedingung warten
      ldi   _ACCA,USITWI.USISIE+USITWI.USIWM1+USITWI.USICS1
      out   USICR,_ACCA                                  ;TWI-Mode, externer Takt, Start-Interrupt
      ldi   _ACCA,USITWI.USISIF+USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC
      out   USISR,_ACCA                                  ;Flags lschen
      rjmp  usiovf_end
usiovf_addr:
      andi   _ACCA,001H                                  ;Richtung erkennen
      breq   usiovf_sendack                              ;Empfang, ACK senden
      lds   _ACCB,USITWI.TxCount                         ;Sendezeichenzhler
      tst   _ACCB                                        ;0 ?
      breq  usiovf_startcond                             ;warten auf neue Startbedingung
      ldi   _ACCA,USITWI.usSENDDATA                      ;beim nchsten Interrupt Daten zum Master senden
      rjmp  usiovf_send
usiovf_sendack:
      ldi   _ACCA,USITWI.usREQDATA                       ;ja, beim nchsten Interrupt Daten vom Master empfangen
usiovf_send:
      clr   _ACCB                                        ;ACK senden
      out   USIDR,_ACCB                                  ;ACK/NAK senden
      sbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Ausgang
      ldi   _ACCB,USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC+00EH
      out   USISR,_ACCB                                  ;Flags lschen auer Startbedingung, Zhler fr 1 Bit
      rjmp  usiovf_endstatus
usiovf_next1:
      cpi   _ACCA,USITWI.usCHECKREPLYSEND                ;teste auf ACK ?
      brne  usiovf_next2                                 ;nein
      in    _ACCA,USIDR                                  ;ACK/NAK holen
      tst   _ACCA                                        ;ACK ?
      brne  usiovf_startcond                             ;nein, auf neue Startbedingung warten
      rjmp  usiovf_senddata                              ;ja, nchste Daten senden
usiovf_next2:
      cpi   _ACCA,USITWI.usSENDDATA                      ;Daten an Master senden ?
      brne  usiovf_next3                                 ;nein
usiovf_senddata:
      lds   _ACCA,USITWI.TxCount                         ;Sendezeichenzhler
      tst   _ACCA                                        ;0 ?
      breq  usiovf_txok                                  ;$00 als Dummy senden
      dec   _ACCA                                        ;erniedrigen
      sts   USITWI.TxCount,_ACCA
      lds   _ACCB,USITWI.iTxRdPtr                        ;Lesezeiger
      clr   _ACCA
      ldi   _ACCCLO,USITWI.oTxBuffer AND 0FFH
      ldi   _ACCCHI,USITWI.oTxBuffer SHRB 8              ;Pufferanfang
      add   _ACCCLO,_ACCB
      adc   _ACCCHI,_ACCA                                ;Leseposition
      ld    _ACCA,Z                                      ;lesen
      inc   _ACCB                                        ;Zeiger weiterrcken
      cpi   _ACCB,TxBuffLen                              ;Pufferende ?
      brne  usiovf_txok                                  ;nein
      clr   _ACCB                                        ;wieder auf Anfang
usiovf_txok:
      out   USIDR,_ACCA                                  ;Byte ausgeben
      sts   USITWI.iTxRdPtr,_ACCB                        ;Pufferzeiger merken
      ldi   _ACCA,USITWI.usREQREPLYSEND                  ;beim nchster Interrupt ACK vom Master empfangen
      sbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Ausgang
      ldi   _ACCB,USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC
      out   USISR,_ACCB                                  ;Flags lschen auer Startbedingung, Zhler auf 0
      rjmp  usiovf_endstatus
usiovf_next3:
      cpi   _ACCA,USITWI.usREQREPLYSEND                  ;auf ACK vom Master warten ?
      brne  usiovf_next4                                 ;nein
      ldi   _ACCA,USITWI.usCHECKREPLYSEND                ;beim nchsten Interrupt ACK vom Master auswerten
      cbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Eingang
      clr   _ACCB
      out   USIDR,_ACCB
      ldi   _ACCB,USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC+00EH
      out   USISR,_ACCB                                  ;Flags lschen auer Startbedingung, Zhler auf 0
      rjmp  usiovf_endstatus
usiovf_next4:
      cpi   _ACCA,USITWI.usREQDATA                       ;auf Daten vom Master warten ?
      brne  usiovf_next5                                 ;nein
      ldi   _ACCA,USITWI.usGETDATAACK                    ;beim nchsten Interrupt Daten lesen und ACK senden
      cbi   USITWI.USIDDR,USITWI.SDAPin                  ;SDA als Eingang
      ldi   _ACCB,USITWI.USIOIF+USITWI.USIPF+USITWI.USIDC
      out   USISR,_ACCB                                  ;Flags lschen auer Startbedingung, Zhler auf 0
      rjmp  usiovf_endstatus
usiovf_next5:
      cpi   _ACCA,USITWI.usGETDATAACK                    ;ACK an Master senden ?
      brne  usiovf_end                                   ;nein
      lds   _ACCA,USITWI.RxCount                         ;Zeichenzhler
      inc   _ACCA                                        ;erhhen
      cpi   _ACCA,RxBuffLen                              ;Maximum
      brne  usiovf_buffok                                ;nein
      rjmp  usiovf_startcond                             ;ja, kein ACK
usiovf_buffok:
      sts   USITWI.RxCount,_ACCA                         ;merken
      lds   _ACCA,USITWI.iRxWrPtr                        ;Pufferzeiger holen
      ldi   _ACCCLO,USITWI.oRxBuffer AND 0FFH
      ldi   _ACCCHI,USITWI.oRxBuffer SHRB 8              ;Pufferanfang
      clr   _ACCB
      add   _ACCCLO,_ACCA
      adc   _ACCCHI,_ACCB                                ;Schreibposition
      lds   _ACCB,USIDR                                  ;Byte lesen
      st    Z,_ACCB                                      ;in Puffer
      inc   _ACCA                                        ;Zeiger weiterrcken
      cpi   _ACCA,RxBuffLen                              ;Pufferende ?
      brne  usiovf_rxok                                  ;nein
      clr   _ACCA                                        ;wieder auf Anfang
usiovf_rxok:
      sts   USITWI.iRxWrPtr,_ACCA                        ;Pufferzeiger merken
      rjmp  usiovf_sendack                               ;ACK senden
usiovf_endstatus:
      sts   USITWI.oTWIState,_ACCA;                      ;neuen Zustand merken
usiovf_end:
    endasm;
  end;

end.


