{*************************************************************************************************}
{*************************************************************************************************}
{**                                                                                             **}
{**                              Serielle Kommunikation ber USI                                **}
{**                                                                                             **}
{*************************************************************************************************}
{*************************************************************************************************}

Unit USIUART;

interface

{*************************************************************************************************}
{ Type Declarations                                                                               }
{*************************************************************************************************}
  type
    TBaudRate = (br2400,br4800,br9600,br14400,br19200,br38400);

{*************************************************************************************************}
{ Var Declarations                                                                                }
{*************************************************************************************************}
{$IDATA}
  var
    RxCount: byte;
    TxCount: byte;                                       //Pufferfllstnde

{*************************************************************************************************}
{ Functions/Procedures                                                                            }
{*************************************************************************************************}
  procedure SerInit;                                     //UART initialisieren
  procedure SetBaud(Baudrate: TBaudRate);                //Baudrate setzen
  function SerStat: boolean;                             //Abfrage ob Empfangsbyte verfgbar sind
  function SerInp: byte;                                 //Empfangsbyte lesen
  procedure SerOut(Data: byte);                          //Daten senden

implementation
{*************************************************************************************************}
{ Const Declarations                                                                              }
{*************************************************************************************************}
  const
    _TX: byte = 0;                                       //Sendebetrieb aktiv
    _TX1: byte = 1;                                      //Sendebetrieb, erste Hlfte
    _RX: byte = 2;                                       //Empfangsbetrieb aktiv
    TX: byte = 1 shl _TX;
    TX1: byte = 1 shl _TX1;
    RX: byte = 1 shl _RX;

    _BR2400: word = loword(PROCCLOCK div longword(2400));
    _BR4800: word = loword(PROCCLOCK div longword(4800));
    _BR9600: word = loword(PROCCLOCK div longword(9600));
    _BR14400: word = loword(PROCCLOCK div longword(14400));
    _BR19200: word = loword(PROCCLOCK div longword(19200));
    _BR38400: word = loword(PROCCLOCK div longword(38400));

{*************************************************************************************************}
{ Var Declarations                                                                                }
{*************************************************************************************************}
  var
{$IDATA}
    iFlags: byte;                                        //Steuerflags
    iPrescaler: byte;                                    //Vorteilerwert
    iCounter: byte;                                      //Zhlerstartwert Empfang
    iTxByte: byte;                                       //aktuelles Sendebyte
    oRxBuffer: array[0..RxBuffLen-1] of byte;            //Empfangspuffer
    oTxBuffer: array[0..TxBuffLen-1] of byte;            //Sendepuffer
    iRxWrPtr: byte;                                      //Schreibzeiger Empfangspuffer
    iRxRdPtr: byte;                                      //Lesezeiger Empfangspuffer
    iTxWrPtr: byte;                                      //Schreibzeiger Sendepuffer
    iTxRdPtr: byte;                                      //Lesezeiger Sendepuffer

{*************************************************************************************************}
{ Functions/Procedures                                                                            }
{*************************************************************************************************}
{$IDATA}
{-------------------------------------------------------------------------------------------------}
{  USI-UART Senden vorbereiten                                                                    }
{-------------------------------------------------------------------------------------------------}
  procedure USIInitTxUART;
  begin
    asm;
      cli                                                ;Interrupt sperren
      lds   _ACCB,USIUART.iFlags                         ;Flags holen
      in    _ACCA,GIMSK
      andi  _ACCA,0EFH
      out   GIMSK,_ACCA                                  ;PinChange-Interrupt aus
      ldi   _ACCA,054H
      out   USICR,_ACCA                                  ;3-Draht-Mode, Takt von Timer 0, Interrupt bei Zhlerberlauf
      ser   _ACCA
      out   USIDR,_ACCA                                  ;Port auf Ruhepegel
      out   USISR,_ACCA                                  ;Interrupts lschen, Zhler auf 15
      in    _ACCA,DDRB
      ori   _ACCA,002H
      out   DDRB,_ACCA                                   ;Portpin als Ausgang
      clr   _ACCA
      out   TCNT0,_ACCA                                  ;Timer 0 lschen
      lds   _ACCA,USIUART.iPrescaler
      out   TCCR0B,_ACCA                                 ;Timer 0 starten
      sbr   _ACCB,USIUART.TX+USIUART.TX1                 ;senden, erste Hlfte
      sts   USIUART.iFlags,_ACCB                         ;Flags speichern
      sbrc  Flags,IntFlag
      sei                                                ;Interrupt freigeben
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  USI-UART Empfang vorbereiten                                                                   }
{-------------------------------------------------------------------------------------------------}
  procedure USIInitRxUART;
  begin
    asm;
      cli                                                ;Interrupt sperren
      clr   _ACCA
      out   TCCR0B,_ACCA                                 ;Timer 0 anhalten
      out   USICR,_ACCA                                  ;USI aus
      in    _ACCA,DDRB
      andi  _ACCA,0FDH
      out   DDRB,_ACCA                                   ;Portpin als Eingang
      ldi   _ACCA,010H
      out   GIFR,_ACCA                                   ;PinChange-Interrupt lschen
      in    _ACCA,GIMSK
      ori   _ACCA,010H
      out   GIMSK,_ACCA                                  ;PinChange-Interrupt ein
      sbrc  Flags,IntFlag
      sei                                                ;Interrupt freigeben
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  UART initialisieren                                                                            }
{-------------------------------------------------------------------------------------------------}
  procedure SerInit;
  begin
    iFlags := 0;                                         //Flags lschen
    iRxWrPtr := 0;
    iRxRdPtr := 0;
    iTxWrPtr := 0;
    iTxRdPtr := 0;                                       //Zeiger auf Pufferanfang
    RxCount := 0;                                        //Empfangszhler auf 0
    TxCount := 0;                                        //Sendezhler auf 0
    PCMSK0 := %00000000;
    PCMSK1 := %00000001;                                 //PinChange-Interrupt auf PB0
    TCCR0A := %00000001;                                 //Timer 0 Clear on Compare
    TCCR0B := %00000000;                                 //Timer 0 angehalten
    SetBaud(br9600);                                     //Baudrate setzen
    USIInitRxUART;                                       //Empfang starten
  end;
{-------------------------------------------------------------------------------------------------}
{  Baudrate setzen                                                                                }
{-------------------------------------------------------------------------------------------------}
{$NOFRAME}
  procedure SetBaud(Baudrate: TBaudRate);
  begin
    asm;
      lsl       _ACCA
      ldi       _ACCCLO,setbaud_tab AND 0FFH
      ldi       _ACCCHI,setbaud_tab SHRB 8               ;Z zeigt auf Tabelle
      add       _ACCCLO,_ACCA
      clr       _ACCA
      adc       _ACCCHI,_ACCA                            ;Offset einrechnen
      ldi       _ACCA,1
      lpm       _ACCBLO,Z+                               ;Tabelle lesen
      lpm       _ACCBHI,Z
setbaud_test:
      tst       _ACCBHI                                  ;Zhler hat nur 8 Bit
      breq      setbaud_ok
      asr       _ACCBHI
      ror       _ACCBLO
      asr       _ACCBHI
      ror       _ACCBLO
      asr       _ACCBHI
      ror       _ACCBLO                                  ;Division durch 8
      inc       _ACCA                                    ;Vorteiler erhhen
      rjmp      setbaud_test:
setbaud_ok:
      dec       _ACCBLO
      out       OCR0A,_ACCBLO                            ;Zhlweite setzen
      ror       _ACCBLO
      andi      _ACCBLO,07FH
      sts       USIUART.iCounter,_ACCBLO                 ;Startwert merken
      sts       USIUART.iPrescaler,_ACCA                 ;Vorteiler merken
      rjmp      setbaud_end
setbaud_tab:
      .WORD     USIUART._BR2400
      .WORD     USIUART._BR4800
      .WORD     USIUART._BR9600
      .WORD     USIUART._BR14400
      .WORD     USIUART._BR19200
      .WORD     USIUART._BR38400
setbaud_end:
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Abfragen, ob Zeichen empfangen wurde                                                           }
{-------------------------------------------------------------------------------------------------}
{$NORETURNCHECK}
  function SerStat: boolean;
  begin
    asm;
      clr   _ACCA
      lds   _ACCB,USIUART.RxCount                        ;Zeichen im Lesepuffer ?
      tst   _ACCB
      breq  serstat_end                                  ;nein
      ser   _ACCA
serstat_end:
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Zeichen aus Empfangspuffer lesen                                                               }
{-------------------------------------------------------------------------------------------------}
{$NORETURNCHECK}
  function SerInp: byte;
  begin
    asm;
serinp_loop:
      lds   _ACCA,USIUART.RxCount                        ;Zeichen im Lesepuffer ?
      tst   _ACCA
      breq  serinp_loop                                  ;nein, warten
      cli                                                ;Interrupts sperren
      lds   _ACCB,USIUART.iRxRdPtr                       ;Lesezeiger
      clr   _ACCA
      ldi   _ACCCLO,USIUART.oRxBuffer AND 0FFH
      ldi   _ACCCHI,USIUART.oRxBuffer SHRB 8             ;Pufferanfang
      add   _ACCCLO,_ACCB
      adc   _ACCCHI,_ACCA                                ;Leseposition
      ld    _ACCA,Z                                      ;aus Puffer
      ldi   _ACCCHI,8                                    ;8 Bit verarbeiten
serinp_rot:
      asr   _ACCA
      rol   _ACCCLO                                      ;Bit umrotieren
      dec   _ACCCHI                                      ;fertig ?
      brne  serinp_rot                                   ;nein
      mov   _ACCA,_ACCCLO                                ;wieder nach ACCA
      inc   _ACCB                                        ;Zeiger weiterrcken
      cpi   _ACCB,RxBuffLen                              ;Pufferende ?
      brne  serinp_ok                                    ;nein
      clr   _ACCB                                        ;wieder auf Anfang
serinp_ok:
      sts   USIUART.iRxRdPtr,_ACCB                       ;Pufferzeiger merken
      lds   _ACCB,USIUART.RxCount                        ;Zeichenzhler
      dec   _ACCB                                        ;erniedrigen
      cpi   _ACCB,0FFH                                   ;Unterlauf
      breq  serinp_end                                   ;ja, verhindern
      sts   USIUART.RxCount,_ACCB                        ;merken
serinp_end:
      sbrc  Flags,IntFlag
      sei                                                ;Interrupts freigeben
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{  Zeichen in Sendepuffer schreiben                                                               }
{-------------------------------------------------------------------------------------------------}
{$DEVICE}
  procedure SerOut(Data: byte);
  begin
    asm;
serout_loop:
      lds   _ACCB,USIUART.TxCount                        ;Fllstand Sendepuffer
      cpi   _ACCB,TxBuffLen                              ;voll ?
      breq  serout_loop                                  ;ja, warten
      ldi   _ACCCHI,8                                    ;8 Bit verarbeiten
serout_rot:
      asr   _ACCA
      rol   _ACCCLO                                      ;Bit umrotieren
      dec   _ACCCHI                                      ;fertig ?
      brne  serout_rot                                   ;nein
      mov   _ACCA,_ACCCLO                                ;wieder nach ACCA
      cli                                                ;Interrupts sperren
      lds   _ACCB,USIUART.iTxWrPtr                       ;Schreibzeiger
      ldi   _ACCCLO,USIUART.oTxBuffer AND 0FFH
      ldi   _ACCCHI,USIUART.oTxBuffer SHRB 8             ;Pufferanfang
      add   _ACCCLO,_ACCB
      clr   _ACCB
      adc   _ACCCHI,_ACCB                                ;Schreibposition
      st    Z,_ACCA                                      ;eintragen
      lds   _ACCB,USIUART.TxCount                        ;Fllstand Sendepuffer
      inc   _ACCB                                        ;+1
      sts   USIUART.TxCount,_ACCB
      lds   _ACCB,USIUART.iTxWrPtr                       ;Schreibzeiger
      inc   _ACCB                                        ;weiterrcken
      cpi   _ACCB,TxBuffLen                              ;Pufferende ?
      brne  serout_ok                                    ;nein
      clr   _ACCB                                        ;wieder auf Anfang
serout_ok:
      sts   USIUART.iTxWrPtr,_ACCB                       ;Pufferzeiger merken
      lds   _ACCB,USIUART.iFlags                         ;Flags holen
      sbrc  _ACCB,USIUART._TX                            ;senden ?
      rjmp  serout_end                                   ;ja, alles ok
serout_wait:
      lds   _ACCB,USIUART.iFlags                         ;Flags holen
      sbrc  _ACCB,USIUART._RX                            ;Empfang luft
      rjmp  serout_wait                                  ;ja, warten
      rcall USIUART.USIInitTxUART                        ;UART auf Senden umschalten
serout_end:
      sbrc  Flags,IntFlag
      sei                                                ;Interrupts freigeben
    endasm;
  end;

{*************************************************************************************************}
{ Interrupt-Routines                                                                              }
{*************************************************************************************************}
{$IDATA}
{-------------------------------------------------------------------------------------------------}
{ Zhler im USI bergelaufen                                                                      }
{-------------------------------------------------------------------------------------------------}
{$NOSAVE}
  interrupt USIOVF;
  begin
    asm;
      lds   _ACCB,USIUART.iFlags                         ;Flags holen
      sbrs  _ACCB,USIUART._TX                            ;senden ?
      rjmp  usiovf_rx                                    ;nein
      ldi   _ACCA,0FBH
      out   USISR,_ACCA                                  ;Interrupts lschen, Zhler auf 11
      sbrc  _ACCB,USIUART._TX1                           ;1. Hlfte ?
      rjmp  usiovf_tx1                                   ;ja
      sbr   _ACCB,USIUART.TX1                            ;folgt wieder 1. Hlfte
      lds   _ACCA,USIUART.iTxByte                        ;Sendedaten holen
      rol   _ACCA
      rol   _ACCA
      rol   _ACCA
      ori   _ACCA,007H                                   ;Stopbyte und Ruhepegel
      out   USIDR,_ACCA                                  ;ausgeben
      rjmp  usiovf_end
usiovf_tx1:
      lds   _ACCA,USIUART.TxCount                        ;Fllstand Sendepuffer
      tst   _ACCA                                        ;leer ?
      brne  usiovf_txdata
      cbr   _ACCB,USIUART.TX+USIUART.TX1                 ;senden beendet
      sts   USIUART.iFlags,_ACCB                         ;Flags speichern
      clr   _ACCA
      out   TCCR0B,_ACCA                                 ;Timer 0 anhalten
      out   USICR,_ACCA                                  ;USI aus
      in    _ACCA,DDRB
      andi  _ACCA,0FDH
      out   DDRB,_ACCA                                   ;Portpin als Eingang
      ldi   _ACCA,010H
      out   GIFR,_ACCA                                   ;PinChange-Interrupt lschen
      in    _ACCA,GIMSK
      ori   _ACCA,010H
      out   GIMSK,_ACCA                                  ;PinChange-Interrupt ein
      rjmp  usiovf_end
usiovf_txdata:
      push  _ACCB
      lds   _ACCB,USIUART.iTxRdPtr                       ;Lesezeiger
      clr   _ACCA
      ldi   _ACCCLO,USIUART.oTxBuffer AND 0FFH
      ldi   _ACCCHI,USIUART.oTxBuffer SHRB 8             ;Pufferanfang
      add   _ACCCLO,_ACCB
      adc   _ACCCHI,_ACCA                                ;Leseposition
      ld    _ACCA,Z                                      ;lesen
      sts   USIUART.iTxByte,_ACCA                        ;merken
      ror   _ACCA
      ror   _ACCA
      andi  _ACCA,03FH
      ori   _ACCA,080H                                   ;Ruhepegel und Startbyte zufgen
      out   USIDR,_ACCA                                  ;ausgeben
      lds   _ACCA,USIUART.TxCount                        ;Sendezeichenzhler
      dec   _ACCA                                        ;erniedrigen
      sts   USIUART.TxCount,_ACCA
      inc   _ACCB                                        ;Zeiger weiterrcken
      cpi   _ACCB,TxBuffLen                              ;Pufferende ?
      brne  udre_ok                                      ;nein
      clr   _ACCB                                        ;wieder auf Anfang
udre_ok:
      sts   USIUART.iTxRdPtr,_ACCB                       ;Pufferzeiger merken
      pop   _ACCB
      cbr   _ACCB,USIUART.TX1                             ;folgt 2. Hlfte
      rjmp  usiovf_end
usiovf_rx:
      sbrs  _ACCB,USIUART._RX                            ;empfangen ?
      rjmp  usiovf_end                                   ;nein
      cbr   _ACCB,USIUART.RX                             ;Empfang beendet
      clr   _ACCA
      out   TCCR0B,_ACCA                                 ;Timer 0 anhalten
      out   USICR,_ACCA                                  ;USI aus
      lds   _ACCA,USIUART.RxCount                        ;Zeichenzhler
      inc   _ACCA                                        ;erhhen
      cpi   _ACCA,RxBuffLen                              ;Maximum
      breq  usiovf_rxfull                                ;ja
      sts   USIUART.RxCount,_ACCA                        ;merken
      push  _ACCB
      lds   _ACCA,USIUART.iRxWrPtr                       ;Pufferzeiger holen
      ldi   _ACCCLO,USIUART.oRxBuffer AND 0FFH
      ldi   _ACCCHI,USIUART.oRxBuffer SHRB 8             ;Pufferanfang
      clr   _ACCB
      add   _ACCCLO,_ACCA
      adc   _ACCCHI,_ACCB                                ;Schreibposition
      lds   _ACCB,USIDR                                  ;Byte von UART
      st    Z,_ACCB                                      ;in Puffer
      inc   _ACCA                                        ;Zeiger weiterrcken
      cpi   _ACCA,RxBuffLen                              ;Pufferende ?
      brne  usiovf_rxok                                  ;nein
      clr   _ACCA                                        ;wieder auf Anfang
usiovf_rxok:
      sts   USIUART.iRxWrPtr,_ACCA                       ;Pufferzeiger merken
      pop   _ACCB
usiovf_rxfull:
      ldi   _ACCA,0FFH
      out   GIFR,_ACCA                                   ;PinChange-Interrupt lschen
      in    _ACCA,GIMSK
      ori   _ACCA,010H
      out   GIMSK,_ACCA                                  ;PinChange-Interrupt ein
usiovf_end:
      sts   USIUART.iFlags,_ACCB                         ;Flags speichern
    endasm;
  end;
{-------------------------------------------------------------------------------------------------}
{ Pegelwechsel am Empfangspin                                                                     }
{-------------------------------------------------------------------------------------------------}
{$NOSAVE}
  interrupt PCINT0;
  begin
    asm;
      lds   _ACCB,USIUART.iFlags                         ;Flags holen
      sbic  PINB,0                                       ;PA0 auf H ?
      rjmp  pcint0_end                                   ;ja, uninteressant
      in    _ACCA,TCCR0B
      ori   _ACCA,008H                                   ;Timer 0 Vorteiler lschen
      out   TCCR0B,_ACCA
      lds   _ACCA,USIUART.iCounter
      out   TCNT0,_ACCA                                  ;Timer 0 Startwert setzen
      lds   _ACCA,USIUART.iPrescaler
      out   TCCR0B,_ACCA                                 ;Timer 0 starten
      ldi   _ACCA,054H
      out   USICR,_ACCA                                  ;3-Draht-Mode, Takt von Timer 0, Interrupt bei Zhlerberlauf
      ldi   _ACCA,0F7H
      out   USISR,_ACCA                                  ;Interrupts lschen, Zhler auf 7
      in    _ACCA,GIMSK
      andi  _ACCA,0EFH
      out   GIMSK,_ACCA                                  ;PinChange-Interrupt aus
      sbr   _ACCB,USIUART.RX                             ;Empfang luft
      sts   USIUART.iFlags,_ACCB                         ;Flags speichern
pcint0_end:
    endasm;
  end;

end.

